@@ -1,22 +1,17 @@
-/* Configuration file parser -*- tab-width: 4; -*-
+/*
+ * Copyright (c) 2002,2003,2007 Martin Hedenfalk <martin@bzero.se>
  *
- * $Id: confuse.c,v 1.27 2004/10/17 16:57:00 mhe Exp $
+ * Permission to use, copy, modify, and distribute this software for any
+ * purpose with or without fee is hereby granted, provided that the above
+ * copyright notice and this permission notice appear in all copies.
  *
- * Copyright (c) 2002-2003, Martin Hedenfalk <mhe@home.se>
- *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public
- * License as published by the Free Software Foundation; either
- * version 2.1 of the License, or (at your option) any later version.
- *
- * This library is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Lesser General Public License for more details.
- *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
+ * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
+ * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
+ * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
+ * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
+ * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
+ * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
+ * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -65,8 +60,8 @@ extern char *cfg_qstring;
 char *cfg_yylval = 0;
 
 const char confuse_version[] = PACKAGE_VERSION;
-const char confuse_copyright[] = PACKAGE_STRING" by Martin Hedenfalk <mhe@home.se>";
-const char confuse_author[] = "Martin Hedenfalk <mhe@home.se>";
+const char confuse_copyright[] = PACKAGE_STRING" by Martin Hedenfalk <martin@bzero.se>";
+const char confuse_author[] = "Martin Hedenfalk <martin@bzero.se>";
 
 static int cfg_parse_internal(cfg_t *cfg, int level,
                               int force_state, cfg_opt_t *force_opt);
@@ -87,7 +82,8 @@ static char *strdup(const char *s)
     if(s == 0 || *s == 0)
         return 0;
 
-    r = (char *)malloc(strlen(s) + 1);
+    r = malloc(strlen(s) + 1);
+    assert(r);
     strcpy(r, s);
     return r;
 }
@@ -102,7 +98,8 @@ static char *strndup(const char *s, size_t n)
     if(s == 0)
         return 0;
 
-    r = (char *)malloc(n + 1);
+    r = malloc(n + 1);
+    assert(r);
     strncpy(r, s, n);
     r[n] = 0;
     return r;
@@ -110,25 +107,28 @@ static char *strndup(const char *s, size_t n)
 #endif
 
 #ifndef HAVE_STRCASECMP
-/* Implementation from GNU glibc 2.3.1
- * Copyright (C) 1991,1992,1995,1996,1997,2001,2002
- *   Free Software Foundation, Inc.
- * License: GNU LGPL
- */
-static int strcasecmp(const char *a, const char *b)
+int strcasecmp(const char *s1, const char *s2)
 {
-    int r;
+    assert(s1);
+    assert(s2);
 
-    if(a == b)
-        return 0;
-
-    while((r = tolower(*a) - tolower(*b++)) == 0)
+    while(*s1)
     {
-        if(*a++ == '\0')
-            break;
+	int c1 = tolower(*(const unsigned char *)s1);
+	int c2 = tolower(*(const unsigned char *)s2);
+	if(c1 < c2)
+	    return -1;
+	if(c1 > c2)
+	    return +1;
+
+	++s1;
+	++s2;
     }
 
-    return r;
+    if(*s2 != 0)
+	return -1;
+
+    return 0;
 }
 #endif
 
@@ -369,15 +369,15 @@ DLLIMPORT cfg_t *cfg_getsec(cfg_t *cfg, const char *name)
 
 static cfg_value_t *cfg_addval(cfg_opt_t *opt)
 {
-    opt->values = (cfg_value_t **)realloc(opt->values,
+    opt->values = realloc(opt->values,
                                   (opt->nvalues+1) * sizeof(cfg_value_t *));
     assert(opt->values);
-    opt->values[opt->nvalues] = (cfg_value_t *)malloc(sizeof(cfg_value_t));
+    opt->values[opt->nvalues] = malloc(sizeof(cfg_value_t));
     memset(opt->values[opt->nvalues], 0, sizeof(cfg_value_t));
     return opt->values[opt->nvalues++];
 }
 
-int cfg_numopts(cfg_opt_t *opts)
+DLLIMPORT int cfg_numopts(cfg_opt_t *opts)
 {
     int n;
 
@@ -392,8 +392,7 @@ static cfg_opt_t *cfg_dupopt_array(cfg_opt_t *opts)
     cfg_opt_t *dupopts;
     int n = cfg_numopts(opts);
 
-    dupopts = (cfg_opt_t *)malloc((n+1) * sizeof(cfg_opt_t));
-    memset(dupopts, 0, (n+1) * sizeof(cfg_opt_t));
+    dupopts = calloc(n+1, sizeof(cfg_opt_t));
     memcpy(dupopts, opts, n * sizeof(cfg_opt_t));
 
     for(i = 0; i < n; i++)
@@ -565,9 +564,13 @@ static cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, char *value)
             {
                 unsigned int i;
 
-                /* check if there already is a section with the same title
+                /* Check if there already is a section with the same title.
                  */
-                assert(value);
+
+		/* Assert that there are either no sections at all, or a
+		 * non-NULL section title. */
+                assert(opt->nvalues == 0 || value);
+
                 for(i = 0; i < opt->nvalues; i++)
                 {
                     cfg_t *sec = opt->values[i]->section;
@@ -582,6 +585,11 @@ static cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, char *value)
                             val = opt->values[i];
                     }
                 }
+                if(val && is_set(CFGF_NO_TITLE_DUPES, opt->flags))
+                {
+                    cfg_error(cfg, _("found duplicate title '%s'"), value);
+                    return 0;
+                }
             }
             if(val == 0)
                 val = cfg_addval(opt);
@@ -662,9 +670,8 @@ static cfg_value_t *cfg_setopt(cfg_t *cfg, cfg_opt_t *opt, char *value)
             if(is_set(CFGF_MULTI, opt->flags) || val->section == 0)
             {
                 cfg_free(val->section);
-                val->section = (cfg_t *)malloc(sizeof(cfg_t));
+                val->section = calloc(1, sizeof(cfg_t));
                 assert(val->section);
-                memset(val->section, 0, sizeof(cfg_t));
                 val->section->name = strdup(opt->name);
                 val->section->opts = cfg_dupopt_array(opt->subopts);
                 val->section->flags = cfg->flags;
@@ -753,13 +760,12 @@ static int call_function(cfg_t *cfg, cfg_opt_t *opt, cfg_opt_t *funcopt)
     /* create a regular argv string vector and call
      * the registered function
      */
-    argv = (const char **)malloc(funcopt->nvalues *
-                                 sizeof(char *));
+    argv = calloc(funcopt->nvalues, sizeof(char *));
     for(i = 0; i < funcopt->nvalues; i++)
         argv[i] = funcopt->values[i]->string;
     ret = (*opt->func)(cfg, opt, funcopt->nvalues, argv);
     cfg_free_value(funcopt);
-    free((char **)argv);
+    free(argv);
     return ret;
 }
 
@@ -893,8 +899,8 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
                     return STATE_ERROR;
                 if(is_set(CFGF_LIST, opt->flags))
                 {
-                    state = 4;
                     ++num_values;
+                    state = 4;
                 }
                 else
                     state = 0;
@@ -903,11 +909,21 @@ static int cfg_parse_internal(cfg_t *cfg, int level,
             case 3: /* expecting an opening brace for a list option */
                 if(tok != '{')
                 {
-                    cfg_error(cfg, _("missing opening brace for option '%s'"),
-                              opt->name);
-                    return STATE_ERROR;
+                    if(tok != CFGT_STR)
+                    {
+                        cfg_error(cfg, _("unexpected token '%s'"), cfg_yylval);
+                        return STATE_ERROR;
+                    }
+
+                    if(cfg_setopt(cfg, opt, cfg_yylval) == 0)
+                        return STATE_ERROR;
+                    if(opt->validcb && (*opt->validcb)(cfg, opt) != 0)
+                        return STATE_ERROR;
+                    ++num_values;
+                    state = 0;
                 }
-                state = 2;
+                else
+                    state = 2;
                 break;
 
             case 4: /* expecting a separator for a list option, or
@@ -1080,9 +1096,8 @@ DLLIMPORT cfg_t *cfg_init(cfg_opt_t *opts, cfg_flag_t flags)
 {
     cfg_t *cfg;
 
-    cfg = (cfg_t *)malloc(sizeof(cfg_t));
+    cfg = calloc(1, sizeof(cfg_t));
     assert(cfg);
-    memset(cfg, 0, sizeof(cfg_t));
 
     cfg->name = strdup("root");
     cfg->opts = cfg_dupopt_array(opts);
@@ -1128,7 +1143,7 @@ DLLIMPORT char *cfg_tilde_expand(const char *filename)
             file = strchr(filename, '/');
             if(file == 0)
                 file = filename + strlen(filename);
-            user = (char *)malloc(file - filename);
+            user = malloc(file - filename);
             strncpy(user, filename + 1, file - filename - 1);
             passwd = getpwnam(user);
             free(user);
@@ -1136,7 +1151,7 @@ DLLIMPORT char *cfg_tilde_expand(const char *filename)
 
         if(passwd)
         {
-            expanded = (char *)malloc(strlen(passwd->pw_dir) + strlen(file) + 1);
+            expanded = malloc(strlen(passwd->pw_dir) + strlen(file) + 1);
             strcpy(expanded, passwd->pw_dir);
             strcat(expanded, file);
         }
@@ -1179,12 +1194,12 @@ static void cfg_free_opt_array(cfg_opt_t *opts)
     for(i = 0; opts[i].name; ++i)
     {
         free(opts[i].name);
-        if(opts[i].type == CFGT_STR)
+        if(opts[i].type == CFGT_FUNC || is_set(CFGF_LIST, opts[i].flags))
+            free(opts[i].def.parsed);
+        else if(opts[i].type == CFGT_STR)
             free(opts[i].def.string);
         else if(opts[i].type == CFGT_SEC)
             cfg_free_opt_array(opts[i].subopts);
-        else if(opts[i].type == CFGT_FUNC || is_set(CFGF_LIST, opts[i].flags))
-            free(opts[i].def.parsed);
     }
     free(opts);
 }
@@ -1397,7 +1412,18 @@ DLLIMPORT void cfg_opt_nprint_var(cfg_opt_t *opt, unsigned int index, FILE *fp)
             break;
         case CFGT_STR:
             str = cfg_opt_getnstr(opt, index);
-            fprintf(fp, "\"%s\"", str ? str : "");
+            fprintf(fp, "\"");
+            while (str && *str)
+	    {
+                if(*str == '"')
+                    fprintf(fp, "\\\"");
+                else if(*str == '\\')
+                    fprintf(fp, "\\\\");
+                else
+                    fprintf(fp, "%c", *str);
+                str++;
+            }
+            fprintf(fp, "\"");
             break;
         case CFGT_BOOL:
             fprintf(fp, "%s", cfg_opt_getnbool(opt, index) ? "true" : "false");
@@ -1565,8 +1591,12 @@ static cfg_opt_t *cfg_getopt_array(cfg_opt_t *rootopts, int cfg_flags, const cha
                 /*fprintf(stderr, "not a section!\n");*/
                 return 0;
             }
-            if(!is_set(CFGF_MULTI, secopt->flags) && (seccfg = cfg_opt_getnsec(secopt, 0)) != 0)
+
+            if(!is_set(CFGF_MULTI, secopt->flags) &&
+	       (seccfg = cfg_opt_getnsec(secopt, 0)) != 0)
+	    {
                 opts = seccfg->opts;
+	    }
             else
                 opts = secopt->subopts;
             if(opts == 0)
